from datetime import datetime

from fhempy.lib import fhem
from fhempy.lib.ble_monitor.blemonitor import BLEmonitor
from fhempy.lib.ble_monitor.bt_helpers import BT_INTERFACES
from fhempy.lib.miscale.body_metrics import bodyMetrics
from fhempy.lib.miscale.body_score import bodyScore

from .. import generic


class miscale(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

        self.blemonitor = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        
        attr_config = {
            "hci_interface": {
                "default": "0",
                "options": ",".join(list(map(str, BT_INTERFACES))),
                "format": "int",
                "help": f"{BT_INTERFACES}",
            },
            "birthday": {
                "default": "",
                "help": "Format: %Y-%m-%d",
            },
            "gender": {
                "default": "",
                "options": "male,female",
            },
            "height": {
                "default": "0",
                "format": "int",
                "help": "In centimeters",
            },
        }
        await self.set_attr_config(attr_config)

        if len(args) != 4:
            return "Usage: define my_miscale fhempy miscale 11:22:33:44:55:66"

        self.mac_addr = args[3]

        self.blemonitor = BLEmonitor.getInstance(self.logger)
        self.blemonitor.register_device(self)

        await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "ready", 1)

    def mac(self):
        return self.mac_addr

    def hci(self):
        return self._attr_hci_interface

    def encryption_key(self):
        return None

    async def received_data(self, data):
        try:
            if "type" in data and data["type"][0:8] != "Mi Scale":
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash,
                    "state",
                    f"detected type {data['type']} is not miscale",
                    1,
                )
                return

            await fhem.readingsBeginUpdate(self.hash)
            for reading in data:
                if reading in ["packet"]:
                    continue
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, reading.replace(" ", "_"), data[reading]
                )

            if "stabilized" in data and data["stabilized"] == 1 and "weight" in data:
                await self.update_scores(data)

            await fhem.readingsEndUpdate(self.hash, 1)
        except Exception:
            self.logger.exception("Failed to update readings")

    async def update_scores(self, data):
        if (
            self._attr_height == 0
            or self._attr_birthday == ""
            or self._attr_gender == ""
        ):
            await fhem.readingsBulkUpdate(
                self.hash, "state", "attr weight/birthday/gender missing"
            )
            return

        weight = data["weight"]
        height = self._attr_height
        age = self.get_age(self._attr_birthday)
        gender = self._attr_gender

        readings = {}
        if data["type"] == "Mi Scale V2":
            if "impedance" not in data:
                return
            impedance = data["impedance"]
            lib = bodyMetrics(weight, height, age, gender, impedance)
            bodyscale = [
                "Obese",
                "Overweight",
                "Thick-set",
                "Lack-exercise",
                "Balanced",
                "Balanced-muscular",
                "Skinny",
                "Balanced-skinny",
                "Skinny-muscular",
            ]
            bmi = lib.getBMI()
            bodyfat = lib.getFatPercentage()
            muscle = lib.getMuscleMass()
            water = lib.getWaterPercentage()
            visceral_fat = lib.getVisceralFat()
            bone = lib.getBoneMass()
            basal_metabolism = lib.getBMR()
            protein = lib.getProteinPercentage()

            readings["bmi"] = "{:.1f}".format(bmi)
            readings["bmr"] = "{:.0f}".format(basal_metabolism)
            readings["visceral_fat"] = "{:.0f}".format(visceral_fat)
            readings["weight_ideal"] = "{:.2f}".format(lib.getIdealWeight())
            readings["bmi_label"] = lib.getBmiLabel()
            readings["lbm_coefficient"] = "{:.1f}".format(lib.getLBMCoefficient())
            readings["bodyfat"] = "{:.1f}".format(bodyfat)
            readings["water"] = "{:.1f}".format(water)
            readings["bone"] = "{:.2f}".format(bone)
            readings["muscle"] = "{:.2f}".format(muscle)
            readings["fat_mass_to_lose_or_gain"] = "{:.2f}".format(
                lib.getFatMassToIdeal()["mass"]
            )
            readings["protein"] = "{:.1f}".format(protein)
            readings["body_type"] = bodyscale[lib.getBodyType()]
            readings["metabolic_age"] = "{:.0f}".format(lib.getMetabolicAge())
            bs = bodyScore(
                age,
                gender,
                height,
                weight,
                bmi,
                bodyfat,
                muscle,
                water,
                visceral_fat,
                bone,
                basal_metabolism,
                protein,
            )
            readings["bodyscore"] = "{:.0f}".format(bs.getBodyScore())
        else:
            lib = bodyMetrics(weight, height, age, gender, 0)
            bmi = lib.getBMI()
            visceral_fat = lib.getVisceralFat()
            basal_metabolism = lib.getBMR()
            readings["bmi"] = "{:.1f}".format(bmi)
            readings["bmr"] = "{:.0f}".format(basal_metabolism)
            readings["visceral_fat"] = "{:.0f}".format(visceral_fat)
            readings["weight_ideal"] = "{:.2f}".format(lib.getIdealWeight())
            readings["bmi_label"] = lib.getBmiLabel()

        for reading_name in readings:
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, reading_name.replace(" ", "_"), readings[reading_name]
            )

    def get_age(self, birthday):
        born = datetime.strptime(birthday, "%Y-%m-%d")
        today = datetime.today()
        return (
            today.year - born.year - ((today.month, today.day) < (born.month, born.day))
        )

    async def set_attr_hci_interface(self, hash):
        self.blemonitor.unregister_device(self)
        self.blemonitor.register_device(self)

    async def Undefine(self, hash):
        if self.blemonitor:
            self.blemonitor.unregister_device(self)
        return await super().Undefine(hash)



import asyncio
import functools
import time

from fhempy.lib import fhem, utils
from fhempy.lib.xiaomi_gateway3_device.devices.base import BaseDevice


class Gateway(BaseDevice):
    def __init__(self, logger, gateway):
        super().__init__(logger, gateway)

    async def initialize(self, device):
        set_config = {
            "start_pairing": {},
            "stop_pairing": {},
            "firmware_update": {"args": ["block"], "options": "block,allow"},
            "start_model_pairing": {"args": ["model"]},
        }
        self.set_set_config(set_config)
        await super().initialize(device)

    async def set_firmware_update(self, hash, params):
        self.create_async_task(self.lock_firmware(params["block"] == "block"))

    async def lock_firmware(self, enable):
        locked = await utils.run_blocking(
            functools.partial(self._gateway.gateway3.lock_firmware, enable)
        )
        await fhem.readingsSingleUpdateIfChanged(
            self.hash, "firmware_lock", str(locked), 1
        )

    async def set_start_model_pairing(self, hash, params):
        model = params["model"]
        self._gateway.gateway3.pair_model = (
            model[:-3] if model.endswith(".v1") else model
        )
        await self.set_start_pairing(hash, params)

    async def set_start_pairing(self, hash, params):
        self._gateway.gateway3.miio.send(
            "miIO.zb_start_provision",
            {
                "dev_type": 0,
                "duration": 60,
                "method": 0,
                "model": "lumi.sensor_switch.v2",
                "pid": 62,
            },
        )

    async def set_stop_pairing(self, hash, params):
        self._gateway.gateway3.miio.send("miIO.zb_end_provision", {"code": -1})

    def update(self, data):
        asyncio.run_coroutine_threadsafe(self.async_update(data), self.loop).result()

    async def async_update(self, data):
        self.last_update = time.time()
        if data is None:
            return

        await fhem.readingsBeginUpdate(self.hash)
        # device is online
        await fhem.readingsBulkUpdateIfChanged(self.hash, "state", "online")
        # update data
        for reading in data:
            if reading == "pairing_start":
                await fhem.readingsBulkUpdateIfChanged(self.hash, "pairing", "on")
            elif reading == "pairing_stop":
                await fhem.readingsBulkUpdateIfChanged(self.hash, "pairing", "off")
            else:
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, reading.replace(" ", "_"), str(data[reading])
                )
        await fhem.readingsEndUpdate(self.hash, 1)



import asyncio
import functools
import logging

import bluepy.btle
import btlewrap
from btlewrap import BluetoothBackendException
from fhempy.lib.generic import FhemModule
from mitemp_bt import mitemp_bt_poller

from .. import fhem, utils


class mitemp(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.hash = None
        self.updateTask = None
        self._poller = None
        self._attr_update_interval = 1200
        self._attr_hci_device = "hci0"
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        
        self._attr_list = {
            "update_interval": {"default": 1200, "format": "int"},
            "hci_device": {"default": "hci0"},
        }
        await self.set_attr_config(self._attr_list)

        if len(args) < 4:
            return "Usage: define mi_temphum fhempy mitemp <MAC>"

        self._address = args[3]
        hash["MAC"] = args[3]
        self.logger.debug(f"Define mitemp: {self._address}")

        self._poller = mitemp_bt_poller.MiTempBtPoller(
            self._address,
            cache_timeout=60,
            adapter=self._attr_hci_device,
            backend=btlewrap.BluepyBackend,
        )

        if self.updateTask:
            self.updateTask.cancel()
        self.updateTask = self.create_async_task(self.update_task())

    async def update_task(self):
        while True:
            self.logger.debug(f"Run update task")
            try:
                await fhem.readingsBeginUpdate(self.hash)
                # name
                name = await utils.run_blocking(functools.partial(self._poller.name))
                await fhem.readingsBulkUpdateIfChanged(self.hash, "name", name)
                # firmware_version
                firmware_version = await utils.run_blocking(
                    functools.partial(self._poller.firmware_version)
                )
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, "firmware", firmware_version
                )
                for param in ("temperature", "humidity", "battery"):
                    # param
                    param_val = await utils.run_blocking(
                        functools.partial(self._poller.parameter_value, param)
                    )
                    await fhem.readingsBulkUpdateIfChanged(self.hash, param, param_val)
                await fhem.readingsBulkUpdateIfChanged(self.hash, "presence", "online")
                await fhem.readingsBulkUpdateIfChanged(self.hash, "state", "online")
                await fhem.readingsEndUpdate(self.hash, 1)
            except Exception:
                self.logger.error(f"Failed to get updates from mitemp {self._address}")
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "presence", "offline", 1
                )
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "state", "offline", 1
                )
            await asyncio.sleep(self._attr_update_interval)

    async def set_attr_update_interval(self, hash):
        await fhem.readingsSingleUpdateIfChanged(
            self.hash, "update_interval", str(self._attr_update_interval), 1
        )
        if self.updateTask:
            self.updateTask.cancel()
        self.updateTask = self.create_async_task(self.update_task())

    async def set_attr_hci_device(self, hash):
        self.logger.debug(f"attr change of hci device")
        self._poller = mitemp_bt_poller.MiTempBtPoller(
            self._address,
            cache_timeout=60,
            adapter=self._attr_hci_device,
            backend=btlewrap.BluepyBackend,
        )



import asyncio

import goodwe as gw

from .. import fhem, generic, utils


class goodwe(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)

        attr_config = {
            "interval": {
                "default": 300,
                "format": "int",
                "help": "Change interval in seconds, default is 300s.",
            }
        }
        await self.set_attr_config(attr_config)

        if len(args) != 4:
            return "Usage: define inverter fhempy goodwe IP"

        self.ip = args[3]

        self.create_async_task(self.update_loop())

    async def update_loop(self):
        self.inverter = None
        while self.inverter is None:
            self.inverter = await gw.connect(self.ip)
            await asyncio.sleep(30)

        while True:
            try:
                runtime_data = await self.inverter.read_runtime_data()
                await self.handle_data(runtime_data)
            except Exception:
                await fhem.readingsSingleUpdate(self.hash, "state", "error", 1)
            await asyncio.sleep(self._attr_interval)

    async def handle_data(self, runtime_data):
        await fhem.readingsBeginUpdate(self.hash)
        try:
            for sensor in self.inverter.sensors():
                if sensor.id_ in runtime_data:
                    reading = utils.gen_reading_name(sensor.name)
                    if sensor.unit:
                        reading += "_" + sensor.unit
                    await fhem.readingsBulkUpdate(
                        self.hash,
                        reading,
                        runtime_data[sensor.id_],
                    )
            await fhem.readingsBulkUpdateIfChanged(self.hash, "state", "connected")
        except Exception:
            self.logger.exception("Failed to update readings")
        await fhem.readingsEndUpdate(self.hash, 1)



import asyncio

from fhempy.lib.generic import FhemModule

from .. import fhem, utils
from .api.client import AsyncSmartmeter


class wienernetze_smartmeter(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) != 5:
            return (
                "Usage: define devname fhempy wienernetze_smartmeter USERNAME PASSWORD"
            )
        self.username = args[3]
        self.password = args[4]

        self._updateloop = self.create_async_task(self.update_loop())

    async def update_loop(self):
        client = AsyncSmartmeter(self.username, self.password)
        await client.refresh_token()
        data = await client.base_information()
        await self.update_readings(data)
        while True:
            try:
                data = await client.consumptions()
                await self.update_readings(data)
                data = await client.meter_readings()
                await self.update_readings(data)
            except Exception:
                self.logger.exception("Failed to update values")
            await asyncio.sleep(3600)

    async def update_readings(self, data):
        flat_data = utils.flatten_json(data)
        await fhem.readingsBeginUpdate(self.hash)
        try:
            for name in flat_data:
                await fhem.readingsBulkUpdateIfChanged(self.hash, name, flat_data[name])
        except Exception:
            self.logger.exception("Failed readings update")
        await fhem.readingsEndUpdate(self.hash, 1)
